/* Benchmark program */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <gmp.h>
#include <gmp-impl.h>

#include "hgcd.h"

#include <time.h>

static double
bench_gcd(mpz_t g, mpz_t a, mpz_t b)
{
  clock_t before;
  clock_t after;
  unsigned count = 0;
  
  before = clock();
  
  do {
    mpz_gcd(g, a, b);
    after = clock();
    count++;
  } while (after < before + CLOCKS_PER_SEC / 10);
  
  return ((double)(after - before)) / CLOCKS_PER_SEC / count;
}

static double
bench_fgcd(mpz_t g, mpz_t a, mpz_t b)
{
  clock_t before;
  clock_t after;
  unsigned count = 0;
  mp_ptr ap;
  mp_ptr bp;
  mp_size_t asize = mpz_size(a);
  mp_size_t bsize = mpz_size(b);
  mpz_t r;

  mpz_init(r);
  mpz_realloc(r, asize + 1);
  
  ap = alloca(sizeof(mp_limb_t) * (asize + bsize + 2));
  bp = ap + asize + 1;

  before = clock();

  do {
    MPN_COPY(ap, a->_mp_d, asize);
    MPN_COPY(bp, b->_mp_d, bsize);
    
    r->_mp_size = mpn_fgcd(r->_mp_d, ap, asize, bp, bsize);
    after = clock();
    count++;
  } while (after < before + CLOCKS_PER_SEC/10);
  
  if (mpz_cmp(g, r) != 0)
    fprintf(stderr, "Bad value from mpz_fgcd\n");

  return ((double)(after - before)) / CLOCKS_PER_SEC / count;
}

static void
header(const char *label)
{
  printf("%s\nlimbs  gcd  fgcd  gcd/fgcd\n", label);
};

static const char *
choose_unit(double *x)
{
  if (*x < 0.001)
    {
      *x *= 1000000;
      return "us";
    }
  else if (*x < 1.0)
    {
      *x *= 1000;
      return "ms";
    }
  return "s";
}

static void
display(unsigned limbs, double gcd_time, double fgcd_time)
{
  const char *gcd_unit;
  const char *fgcd_unit;
  
  double q = gcd_time / fgcd_time;
  double l = log(q) / log(2);
  int pos;

  gcd_unit = choose_unit(&gcd_time);
  fgcd_unit = choose_unit(&fgcd_time);

  printf("%5u %5.2f%s %5.2f%s %5.2f",
	 limbs, gcd_time, gcd_unit, fgcd_time, fgcd_unit,
	 q);

  pos = (l + 5 ) * 60 / 20;
  if (pos >= 0 && pos <= 60)
    {
      int i;
      for (i = 0; i<pos; i++)
	printf(" ");
      printf("*");
    }
  printf("\n");
}

int
main(int argc, char **argv)
{
  mpz_t a;
  mpz_t b;
  mpz_t c;
  mpz_t g;

  unsigned i;
  
  mpz_init(a);
  mpz_init(b);
  mpz_init(c);
  mpz_init(g);
  
  header("Gcd of fibonacci numbers.\n");

  for (i = 10; ; i = i * 3 / 2)
    {
      double gcd_time;
      double fgcd_time;
      unsigned limbs;
      
      mpz_fib2_ui(a, b, i);

      limbs = mpz_size(b);

      gcd_time = bench_gcd(g, a, b);
      fgcd_time = bench_fgcd(g, a, b);

      display(limbs, gcd_time, fgcd_time);
      
      if (gcd_time > 25.0 || fgcd_time > 25.0)
	break;
    }

  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(c);
  mpz_clear(g);

  return 0;
}
